import logging
from typing import List

import pandas as pd

from .tokenizer import Tokenizer, Token

logger = logging.getLogger(__name__)


class Analyzer:
    def __init__(self, tokenizer: Tokenizer, char_filters=[], token_filters=[]):
        """Analyzer Object

        Parameters
        ----------
        tokenizer: Tokenizer
            Tokenizer object
        char_filters : list, optional
            CharFilters list, by default []
        token_filters : list, optional
            TokenFilters list, by default []
        """

        self.tokenizer = tokenizer
        self.char_filters = char_filters
        self.token_filters = token_filters

    def analyze(self, text: str, wakati: bool = False):
        """品詞に分解するメソッド.

        Parameters
        ----------
        text : str
            解析する文字列
        wakati: bool, optional
            Trueにするとsurface部分のみをリストで返すようにする

        Returns
        -------
        トークンのジェネレーター
        """

        if not text:
            logger.info("text is empty!")
            return
        for cfilter in self.char_filters:
            text = cfilter.filter(text)

        tokens: List[Token] = self.tokenizer.tokenize(text)
        for tfilter in self.token_filters:
            tokens = tfilter.filter(tokens)

        if wakati:
            return [token.surface for token in tokens]

        return tokens

    def analyze_with_dataframe(self, text: str):
        """品詞分解したものをpd.DataFrameで返すメソッド

        Parameters
        ----------
        text : str
            解析する文字列

        Returns
        -------
        pd.DataFrame
        """
        if not text:
            logger.info("text is empty!")
            return

        tokens: List[Token] = list(self.analyze(text))
        try:
            columns = list(tokens[0].__dict__.keys())
        except IndexError:
            columns = [
                "surface",
                "part_of_speech",
                "infl_type",
                "infl_form",
                "base_form",
                "reading",
                "phonetic",
            ]
            return pd.DataFrame(columns=columns)

        dataframe = {}
        for col in columns:
            if col != "cache":
                dataframe[col] = [getattr(token, col) for token in tokens]

        return pd.DataFrame(dataframe)



import re
import unicodedata


class CharFilter:
    def filter(self, text):
        return self.apply(text)

    def apply(self, text):
        raise NotImplementedError


class RegexReplaceCharFilter(CharFilter):
    def __init__(self, pat, repl):
        """正規表現に一致する部分を置き換える.

        Parameters
        ----------
        pat : str
            文字列型の正規表現.
        repl : str
            置き換える文字列.
        """
        self.pattern = re.compile(pat)
        self.replacement = repl

    def apply(self, text):
        return re.sub(self.pattern, self.replacement, text)


class UnicodeNormalizeCharFilter(CharFilter):
    """
    ユニコード文字列の正規化を行うフィルター
    """

    def __init__(self, form="NFKC"):
        """

        Parameters
        ----------
        form : str, optional
            正規化の形式. 'NFC', 'NFKC, 'NFD'のどれかを選ぶこと. by default 'NFKC'
        """
        self.form = form

    def apply(self, text):
        return unicodedata.normalize(self.form, text)


class URLReplaceFilter(RegexReplaceCharFilter):
    """ URLを指定された文字列に変換するフィルター """
    def __init__(self, repl=""):
        self.pattern = re.compile(
            r"(https?|ftp)(:\/\/[-_\.!~*\'()a-zA-Z0-9;\/?:\@&=\+\$,%#]+)"
        )
        self.replacement = repl



from typing import Union, List
import re
import subprocess

import MeCab


class Token:
    """
    tokenの情報を内包するクラス
    """

    def __init__(self, word_with_feat: str):
        surface, feature = word_with_feat.split("\t")

        # mecabの仕様で,アルファベットの単語は原型が'*'になり,
        # 読みと発音が省略されるためその対応として以下のように行う.
        split_feature = feature.split(",")
        if len(split_feature) == 7:
            part_of_speech = ",".join(split_feature[:4])
            infl_type = split_feature[4]
            infl_form = split_feature[5]
            base_form = surface
            reading = "*"
            phonetic = "*"
        else:
            (
                part_of_speech,
                infl_type,
                infl_form,
                base_form,
                reading,
                phonetic,
            ) = feature.rsplit(",", 5)

        self.surface = surface
        """surface form (表層形)"""
        self.part_of_speech = part_of_speech
        """part of speech (品詞)"""
        self.infl_type = infl_type
        """terminal form (活用型)"""
        self.infl_form = infl_form
        """stem form (活用形)"""
        self.base_form = base_form
        """base form (基本形)"""
        self.reading = reading
        """"reading (読み)"""
        self.phonetic = phonetic
        """pronounce (発音)"""

        # もとのsurfaceを残しておくもの
        self.cache = surface

    def __str__(self):
        return "%s\t%s,%s,%s,%s,%s,%s" % (
            self.surface,
            self.part_of_speech,
            self.infl_type,
            self.infl_form,
            self.base_form,
            self.reading,
            self.phonetic,
        )


class Tokenizer:
    """
    textの品詞分解を司るクラス
    """

    def __init__(self, use_neologd=False):
        """mecabを使えるようにする

        Parameters
        ----------
        use_neologd : bool, optional
            NEologdを使うかどうかを指定する, by default False
        """
        # get neologd
        cmd = "echo `mecab-config --dicdir`"
        if use_neologd:
            cmd = cmd + '"/mecab-ipadic-neologd"'
        else:
            cmd = cmd + '"/ipadic"'

        # 辞書のパスを取得する
        dict_path = (
            subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
            .stdout.readline()
            .decode()
        )
        dict_path = re.sub(r"\n", "", dict_path)

        self.tokenizer = MeCab.Tagger("-d " + dict_path)

    def tokenize(self, text, wakati=False) -> Union[List[Token]]:
        """トークン化する

        Parameters
        ----------
        text : str
            品詞分解する対象のtext
        wakati : bool, optional
            Trueにするとsurface部分のみをリストで返すようにする, by default False

        Returns
        -------
        トークンのリスト(wakati=False),もしくは文字列のリスト(wakati=True)
        """
        chunk: str = self.tokenizer.parse(text).split("\n")[:-2]
        if wakati:
            tokens = [Token(ch).surface for ch in chunk]
        else:
            tokens = [Token(ch) for ch in chunk]
        return tokens
